package java.util;

import java.util.stream.StreamSupport;
import java.util.stream.IntStream;
import java.io.ObjectInputStream;
import java.io.IOException;
import java.io.ObjectOutputStream;
import java.nio.ByteOrder;
import java.nio.ByteBuffer;
import java.nio.LongBuffer;
import java.io.ObjectStreamField;
import java.io.Serializable;

public class BitSet implements Cloneable, Serializable
{
    private static final int ADDRESS_BITS_PER_WORD = 6;
    private static final int BITS_PER_WORD = 64;
    private static final int BIT_INDEX_MASK = 63;
    private static final long WORD_MASK = -1L;
    private static final ObjectStreamField[] serialPersistentFields;
    private long[] words;
    private transient int wordsInUse;
    private transient boolean sizeIsSticky;
    private static final long serialVersionUID = 7997698588986878753L;
    
    private static int wordIndex(int n) {
        return n >> 6;
    }
    
    private void checkInvariants() {
        assert this.words[this.wordsInUse - 1] != 0L;
        assert this.wordsInUse >= 0 && this.wordsInUse <= this.words.length;
        assert this.words[this.wordsInUse] == 0L;
        return;
    }
    
    private void recalculateWordsInUse() {
        int n;
        for (n = this.wordsInUse - 1; n >= 0 && this.words[n] == 0L; --n) {}
        this.wordsInUse = n + 1;
    }
    
    public BitSet() {
        this.wordsInUse = 0;
        this.sizeIsSticky = false;
        this.initWords(64);
        this.sizeIsSticky = false;
    }
    
    public BitSet(int i) {
        this.wordsInUse = 0;
        this.sizeIsSticky = false;
        if (i < 0) {
            throw new NegativeArraySizeException("nbits < 0: " + i);
        }
        this.initWords(i);
        this.sizeIsSticky = true;
        return;
    }
    
    private void initWords(int n) {
        this.words = new long[wordIndex(n - 1) + 1];
    }
    
    private BitSet(long[] words) {
        this.wordsInUse = 0;
        this.sizeIsSticky = false;
        this.words = words;
        this.wordsInUse = words.length;
        this.checkInvariants();
    }
    
    public static BitSet valueOf(long[] array) {
        int length;
        for (length = array.length; length > 0 && array[length - 1] == 0L; --length) {}
        return new BitSet(Arrays.copyOf(array, length));
    }
    
    public static BitSet valueOf(LongBuffer slice) {
        int remaining;
        for (slice = slice.slice(), remaining = slice.remaining(); remaining > 0 && slice.get(remaining - 1) == 0L; --remaining) {}
        long[] dst = new long[remaining];
        slice.get(dst);
        return new BitSet(dst);
    }
    
    public static BitSet valueOf(byte[] array) {
        return valueOf(ByteBuffer.wrap(array));
    }
    
    public static BitSet valueOf(ByteBuffer order) {
        int remaining;
        for (order = order.slice().order(ByteOrder.LITTLE_ENDIAN), remaining = order.remaining(); remaining > 0 && order.get(remaining - 1) == 0; --remaining) {}
        long[] array = new long[(remaining + 7) / 8];
        order.limit(remaining);
        int n = 0;
        while (order.remaining() >= 8) {
            array[n++] = order.getLong();
        }
        for (int remaining2 = order.remaining(), i = 0; i < remaining2; ++i) {
            int n2 = n;
            array[n2] |= ((long)order.get() & 0xFFL) << 8 * i;
        }
        return new BitSet(array);
    }
    
    public byte[] toByteArray() {
        int wordsInUse = this.wordsInUse;
        if (wordsInUse == 0) {
            return new byte[0];
        }
        int n = 8 * (wordsInUse - 1);
        for (long n2 = this.words[wordsInUse - 1]; n2 != 0L; n2 >>>= 8) {
            ++n;
        }
        byte[] array = new byte[n];
        ByteBuffer order = ByteBuffer.wrap(array).order(ByteOrder.LITTLE_ENDIAN);
        for (int i = 0; i < wordsInUse - 1; ++i) {
            order.putLong(this.words[i]);
        }
        for (long n3 = this.words[wordsInUse - 1]; n3 != 0L; n3 >>>= 8) {
            order.put((byte)(int)(n3 & 0xFFL));
        }
        return array;
    }
    
    public long[] toLongArray() {
        return Arrays.copyOf(this.words, this.wordsInUse);
    }
    
    private void ensureCapacity(int b) {
        if (this.words.length < b) {
            this.words = Arrays.copyOf(this.words, Math.max(2 * this.words.length, b));
            this.sizeIsSticky = false;
        }
    }
    
    private void expandTo(int n) {
        int wordsInUse = n + 1;
        if (this.wordsInUse < wordsInUse) {
            this.ensureCapacity(wordsInUse);
            this.wordsInUse = wordsInUse;
        }
    }
    
    private static void checkRange(int n, int n2) {
        if (n < 0) {
            throw new IndexOutOfBoundsException("fromIndex < 0: " + n);
        }
        if (n2 < 0) {
            throw new IndexOutOfBoundsException("toIndex < 0: " + n2);
        }
        if (n > n2) {
            throw new IndexOutOfBoundsException("fromIndex: " + n + " > toIndex: " + n2);
        }
        return;
    }
    
    public void flip(int i) {
        if (i < 0) {
            throw new IndexOutOfBoundsException("bitIndex < 0: " + i);
        }
        int wordIndex = wordIndex(i);
        this.expandTo(wordIndex);
        long[] words = this.words;
        int n = wordIndex;
        words[n] ^= 1L << i;
        this.recalculateWordsInUse();
        this.checkInvariants();
        return;
    }
    
    public void flip(int n, int n2) {
        checkRange(n, n2);
        if (n == n2) {
            return;
        }
        int wordIndex = wordIndex(n);
        int wordIndex2 = wordIndex(n2 - 1);
        this.expandTo(wordIndex2);
        long n3 = -1L << n;
        long n4 = -1L >>> -n2;
        if (wordIndex == wordIndex2) {
            long[] words4 = this.words;
            int n8 = wordIndex;
            words4[n8] ^= (n3 & n4);
        }
        else {
            long[] words = this.words;
            int n5 = wordIndex;
            words[n5] ^= n3;
            for (int i = wordIndex + 1; i < wordIndex2; ++i) {
                long[] words2 = this.words;
                int n6 = i;
                words2[n6] ^= -1L;
            }
            long[] words3 = this.words;
            int n7 = wordIndex2;
            words3[n7] ^= n4;
        }
        this.recalculateWordsInUse();
        this.checkInvariants();
        return;
    }
    
    public void set(int i) {
        if (i < 0) {
            throw new IndexOutOfBoundsException("bitIndex < 0: " + i);
        }
        int wordIndex = wordIndex(i);
        this.expandTo(wordIndex);
        long[] words = this.words;
        int n = wordIndex;
        words[n] |= 1L << i;
        this.checkInvariants();
        return;
    }
    
    public void set(int n, boolean b) {
        if (b) {
            this.set(n);
        }
        else {
            this.clear(n);
        }
    }
    
    public void set(int n, int n2) {
        checkRange(n, n2);
        if (n == n2) {
            return;
        }
        int wordIndex = wordIndex(n);
        int wordIndex2 = wordIndex(n2 - 1);
        this.expandTo(wordIndex2);
        long n3 = -1L << n;
        long n4 = -1L >>> -n2;
        if (wordIndex == wordIndex2) {
            long[] words3 = this.words;
            int n7 = wordIndex;
            words3[n7] |= (n3 & n4);
        }
        else {
            long[] words = this.words;
            int n5 = wordIndex;
            words[n5] |= n3;
            for (int i = wordIndex + 1; i < wordIndex2; ++i) {
                this.words[i] = -1L;
            }
            long[] words2 = this.words;
            int n6 = wordIndex2;
            words2[n6] |= n4;
        }
        this.checkInvariants();
        return;
    }
    
    public void set(int n, int n2, boolean b) {
        if (b) {
            this.set(n, n2);
        }
        else {
            this.clear(n, n2);
        }
    }
    
    public void clear(int i) {
        if (i < 0) {
            throw new IndexOutOfBoundsException("bitIndex < 0: " + i);
        }
        int wordIndex = wordIndex(i);
        if (wordIndex >= this.wordsInUse) {
            return;
        }
        long[] words = this.words;
        int n = wordIndex;
        words[n] &= ~(1L << i);
        this.recalculateWordsInUse();
        this.checkInvariants();
        return;
    }
    
    public void clear(int n, int length) {
        checkRange(n, length);
        if (n == length) {
            return;
        }
        int wordIndex = wordIndex(n);
        if (wordIndex >= this.wordsInUse) {
            return;
        }
        int wordIndex2 = wordIndex(length - 1);
        if (wordIndex2 >= this.wordsInUse) {
            length = this.length();
            wordIndex2 = this.wordsInUse - 1;
        }
        long n2 = -1L << n;
        long n3 = -1L >>> -length;
        if (wordIndex == wordIndex2) {
            long[] words3 = this.words;
            int n6 = wordIndex;
            words3[n6] &= ~(n2 & n3);
        }
        else {
            long[] words = this.words;
            int n4 = wordIndex;
            words[n4] &= ~n2;
            for (int i = wordIndex + 1; i < wordIndex2; ++i) {
                this.words[i] = 0L;
            }
            long[] words2 = this.words;
            int n5 = wordIndex2;
            words2[n5] &= ~n3;
        }
        this.recalculateWordsInUse();
        this.checkInvariants();
        return;
    }
    
    public void clear() {
        while (this.wordsInUse > 0) {
            this.words[--this.wordsInUse] = 0L;
        }
    }
    
    public boolean get(int i) {
        if (i < 0) {
            throw new IndexOutOfBoundsException("bitIndex < 0: " + i);
        }
        this.checkInvariants();
        int wordIndex = wordIndex(i);
        return wordIndex < this.wordsInUse && (this.words[wordIndex] & 1L << i) != 0x0L;
    }
    
    public BitSet get(int n, int n2) {
        checkRange(n, n2);
        this.checkInvariants();
        int length = this.length();
        if (length <= n || n == n2) {
            return new BitSet(0);
        }
        if (n2 > length) {
            n2 = length;
        }
        BitSet o = new BitSet(n2 - n);
        int wordsInUse = wordIndex(n2 - n - 1) + 1;
        int wordIndex = wordIndex(n);
        int n3 = (n & 0x3F) == 0x0;
        for (int i = 0; i < wordsInUse - 1; ++i, ++wordIndex) {
            o.words[i] = (n3 ? this.words[wordIndex] : (this.words[wordIndex] >>> n | this.words[wordIndex + 1] << -n));
        }
        long n4 = -1L >>> -n2;
        o.words[wordsInUse - 1] = (((n2 - 1 & 0x3F) < (n & 0x3F)) ? (this.words[wordIndex] >>> n | (this.words[wordIndex + 1] & n4) << -n) : ((this.words[wordIndex] & n4) >>> n));
        o.wordsInUse = wordsInUse;
        o.recalculateWordsInUse();
        o.checkInvariants();
        return o;
    }
    
    public int nextSetBit(int i) {
        if (i < 0) {
            throw new IndexOutOfBoundsException("fromIndex < 0: " + i);
        }
        this.checkInvariants();
        int wordIndex = wordIndex(i);
        if (wordIndex >= this.wordsInUse) {
            return -1;
        }
        long j = this.words[wordIndex] & -1L << i;
        while (j == 0L) {
            if (++wordIndex == this.wordsInUse) {
                return -1;
            }
            j = this.words[wordIndex];
        }
        return wordIndex * 64 + Long.numberOfTrailingZeros(j);
    }
    
    public int nextClearBit(int i) {
        if (i < 0) {
            throw new IndexOutOfBoundsException("fromIndex < 0: " + i);
        }
        this.checkInvariants();
        int wordIndex = wordIndex(i);
        if (wordIndex >= this.wordsInUse) {
            return i;
        }
        long j = ~this.words[wordIndex] & -1L << i;
        while (j == 0L) {
            if (++wordIndex == this.wordsInUse) {
                return this.wordsInUse * 64;
            }
            j = ~this.words[wordIndex];
        }
        return wordIndex * 64 + Long.numberOfTrailingZeros(j);
    }
    
    public int previousSetBit(int i) {
        if (i < 0) {
            if (i == -1) {
                return -1;
            }
            throw new IndexOutOfBoundsException("fromIndex < -1: " + i);
        }
        else {
            this.checkInvariants();
            int wordIndex = wordIndex(i);
            if (wordIndex >= this.wordsInUse) {
                return this.length() - 1;
            }
            long j = this.words[wordIndex] & -1L >>> -(i + 1);
            while (j == 0L) {
                if (wordIndex-- == 0) {
                    return -1;
                }
                j = this.words[wordIndex];
            }
            return (wordIndex + 1) * 64 - 1 - Long.numberOfLeadingZeros(j);
        }
    }
    
    public int previousClearBit(int i) {
        if (i < 0) {
            if (i == -1) {
                return -1;
            }
            throw new IndexOutOfBoundsException("fromIndex < -1: " + i);
        }
        else {
            this.checkInvariants();
            int wordIndex = wordIndex(i);
            if (wordIndex >= this.wordsInUse) {
                return i;
            }
            long j = ~this.words[wordIndex] & -1L >>> -(i + 1);
            while (j == 0L) {
                if (wordIndex-- == 0) {
                    return -1;
                }
                j = ~this.words[wordIndex];
            }
            return (wordIndex + 1) * 64 - 1 - Long.numberOfLeadingZeros(j);
        }
    }
    
    public int length() {
        if (this.wordsInUse == 0) {
            return 0;
        }
        return 64 * (this.wordsInUse - 1) + (64 - Long.numberOfLeadingZeros(this.words[this.wordsInUse - 1]));
    }
    
    public boolean isEmpty() {
        return this.wordsInUse == 0;
    }
    
    public boolean intersects(BitSet set) {
        int i = Math.min(this.wordsInUse, set.wordsInUse) - 1;
        while (i >= 0) {
            if ((this.words[i] & set.words[i]) != 0x0L) {
                return true;
            }
            --i;
        }
        return false;
    }
    
    public int cardinality() {
        int n = 0;
        for (int i = 0; i < this.wordsInUse; ++i) {
            n += Long.bitCount(this.words[i]);
        }
        return n;
    }
    
    public void and(BitSet set) {
        if (this == set) {
            return;
        }
        while (this.wordsInUse > set.wordsInUse) {
            this.words[--this.wordsInUse] = 0L;
        }
        for (int i = 0; i < this.wordsInUse; ++i) {
            long[] words = this.words;
            int n = i;
            words[n] &= set.words[i];
        }
        this.recalculateWordsInUse();
        this.checkInvariants();
        return;
    }
    
    public void or(BitSet set) {
        if (this == set) {
            return;
        }
        int min = Math.min(this.wordsInUse, set.wordsInUse);
        if (this.wordsInUse < set.wordsInUse) {
            this.ensureCapacity(set.wordsInUse);
            this.wordsInUse = set.wordsInUse;
        }
        for (int i = 0; i < min; ++i) {
            long[] words = this.words;
            int n = i;
            words[n] |= set.words[i];
        }
        if (min < set.wordsInUse) {
            System.arraycopy(set.words, min, this.words, min, this.wordsInUse - min);
        }
        this.checkInvariants();
        return;
    }
    
    public void xor(BitSet set) {
        int min = Math.min(this.wordsInUse, set.wordsInUse);
        if (this.wordsInUse < set.wordsInUse) {
            this.ensureCapacity(set.wordsInUse);
            this.wordsInUse = set.wordsInUse;
        }
        for (int i = 0; i < min; ++i) {
            long[] words = this.words;
            int n = i;
            words[n] ^= set.words[i];
        }
        if (min < set.wordsInUse) {
            System.arraycopy(set.words, min, this.words, min, set.wordsInUse - min);
        }
        this.recalculateWordsInUse();
        this.checkInvariants();
    }
    
    public void andNot(BitSet set) {
        for (int i = Math.min(this.wordsInUse, set.wordsInUse) - 1; i >= 0; --i) {
            long[] words = this.words;
            int n = i;
            words[n] &= ~set.words[i];
        }
        this.recalculateWordsInUse();
        this.checkInvariants();
    }
    
    @Override
    public int hashCode() {
        long n = 1234L;
        int wordsInUse = this.wordsInUse;
        while (--wordsInUse >= 0) {
            n ^= this.words[wordsInUse] * (long)(wordsInUse + 1);
        }
        return (int)(n >> 32 ^ n);
    }
    
    public int size() {
        return this.words.length * 64;
    }
    
    @Override
    public boolean equals(Object o) {
        if (!(o instanceof BitSet)) {
            return false;
        }
        if (this == o) {
            return true;
        }
        BitSet set = (BitSet)o;
        this.checkInvariants();
        set.checkInvariants();
        if (this.wordsInUse != set.wordsInUse) {
            return false;
        }
        int i = 0;
        while (i < this.wordsInUse) {
            if (this.words[i] != set.words[i]) {
                return false;
            }
            ++i;
        }
        return true;
    }
    
    public Object clone() {
        if (!this.sizeIsSticky) {
            this.trimToSize();
        }
        try {
            BitSet set = (BitSet)super.clone();
            set.words = (long[])this.words.clone();
            set.checkInvariants();
            return set;
        }
        catch (CloneNotSupportedException cause) {
            throw new InternalError(cause);
        }
    }
    
    private void trimToSize() {
        if (this.wordsInUse != this.words.length) {
            this.words = Arrays.copyOf(this.words, this.wordsInUse);
            this.checkInvariants();
        }
    }
    
    private void writeObject(ObjectOutputStream objectOutputStream) throws IOException {
        this.checkInvariants();
        if (!this.sizeIsSticky) {
            this.trimToSize();
        }
        objectOutputStream.putFields().put("bits", this.words);
        objectOutputStream.writeFields();
    }
    
    private void readObject(ObjectInputStream objectInputStream) throws IOException, ClassNotFoundException {
        this.words = (long[])(long[])objectInputStream.readFields().get("bits", null);
        this.wordsInUse = this.words.length;
        this.recalculateWordsInUse();
        this.sizeIsSticky = (this.words.length > 0 && this.words[this.words.length - 1] == 0L);
        this.checkInvariants();
    }
    
    @Override
    public String toString() {
        this.checkInvariants();
        StringBuilder sb = new StringBuilder(6 * ((this.wordsInUse > 128) ? this.cardinality() : (this.wordsInUse * 64)) + 2);
        sb.append('{');
        int n = this.nextSetBit(0);
        if (n != -1) {
            sb.append(n);
            while (++n >= 0) {
                if ((n = this.nextSetBit(n)) < 0) {
                    break;
                }
                do {
                    sb.append(", ").append(n);
                } while (++n != this.nextClearBit(n));
            }
        }
        sb.append('}');
        return sb.toString();
    }
    
    public IntStream stream() {
        class BitSetIterator implements PrimitiveIterator.OfInt
        {
            int next;
            
            BitSetIterator() {
                this.next = BitSet.this.nextSetBit(0);
            }
            
            @Override
            public boolean hasNext() {
                return this.next != -1;
            }
            
            @Override
            public int nextInt() {
                if (this.next != -1) {
                    int next = this.next;
                    this.next = BitSet.this.nextSetBit(this.next + 1);
                    return next;
                }
                throw new NoSuchElementException();
            }
        }
        return StreamSupport.intStream(()/*  */ -> Spliterators.spliterator(new BitSetIterator(), (long)this.cardinality(), 21), 16469, false);
    }
    
    static {
        serialPersistentFields = new ObjectStreamField[] { new ObjectStreamField("bits", long[].class) };
    }
}
